@@ -2,17 +2,17 @@
    Copyright (C) 2000-2025 Free Software Foundation, Inc.
 
    Written by Sergey Poznyakoff
-  
+
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
-  
+
    GNU Radius is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-  
+
    You should have received a copy of the GNU General Public License
    along with GNU Radius.  If not, see <http://www.gnu.org/licenses/>. */
 
@@ -33,44 +33,44 @@ struct radutmp; /* declared in radutmp.h */
    fields to it. */
 typedef struct radiusd_request {
 	grad_request_t *request;
-	
-        /* Saved reply values */
-        int           reply_code;   /* Reply code */
-        grad_avp_t    *reply_pairs; /* Reply pairs */
-        char          *reply_msg;   /* Reply message */
-	                            /* FIXME: should probably be
-			 	       incorporated to reply_pairs
+
+	/* Saved reply values */
+	int           reply_code;   /* Reply code */
+	grad_avp_t    *reply_pairs; /* Reply pairs */
+	char          *reply_msg;   /* Reply message */
+				    /* FIXME: should probably be
+				       incorporated to reply_pairs
 				       at once */
 	/* List of cfg file locations that lead to the decision on this
 	   request */
 	grad_list_t   *locus_list;
-	
-        /* Proxy support fields */
-        grad_realm_t  *realm;       
-        int           validated;     /* Already md5 checked */
+
+	/* Proxy support fields */
+	grad_realm_t  *realm;
+	int           validated;     /* Already md5 checked */
 	int           server_no;
 	int           attempt_no;
-        uint32_t server_id;     /* Proxy ID of the packet */
+	uint32_t server_id;     /* Proxy ID of the packet */
 	char          *remote_user;  /* Remote username */
-        u_char        remote_auth[GRAD_AUTHENTICATOR_LENGTH];
-	                             /* Remote request authenticator */	
-        int           server_code;   /* Reply code from other srv */
-        grad_avp_t    *server_reply; /* Reply from other server */
+	u_char        remote_auth[GRAD_AUTHENTICATOR_LENGTH];
+				     /* Remote request authenticator */
+	int           server_code;   /* Reply code from other srv */
+	grad_avp_t    *server_reply; /* Reply from other server */
 } radiusd_request_t;
 
 enum reload_what {
-        reload_config,
-        reload_all,
-        reload_dict,
-        reload_users,
-        reload_huntgroups,
-        reload_hints,
-        reload_clients, 
-        reload_naslist, 
-        reload_realms,
-        reload_deny,
-        reload_sql,
-        reload_rewrite
+	reload_config,
+	reload_all,
+	reload_dict,
+	reload_users,
+	reload_huntgroups,
+	reload_hints,
+	reload_clients,
+	reload_naslist,
+	reload_realms,
+	reload_deny,
+	reload_sql,
+	reload_rewrite
 };
 
 /* ********************** Request list handling **************************** */
@@ -91,7 +91,7 @@ enum reload_what {
 			       handler and is waiting for it to
 			       become free */
 #define RS_TERMINATED 4     /* TERM has been sent to the handler */
-			      
+
 /* Request comparison results */
 #define RCMP_NE     0      /* Requests not equal */
 #define RCMP_EQ     1      /* Requests equal */
@@ -100,16 +100,16 @@ enum reload_what {
 typedef struct request REQUEST;
 
 struct request {
-        int             type;         /* request type */
-        int             status;       /* request status */
-        time_t          timestamp;    /* when was the request accepted */
-        pid_t           child_id;     /* ID of the handling process */
-        int             code;         /* Child return code if completed */
-        void            *data;        /* Request-specific data */
+	int             type;         /* request type */
+	int             status;       /* request status */
+	time_t          timestamp;    /* when was the request accepted */
+	pid_t           child_id;     /* ID of the handling process */
+	int             code;         /* Child return code if completed */
+	void            *data;        /* Request-specific data */
 	void            *rawdata;     /* Raw data as received from the
-					 socket */  
+					 socket */
 	size_t          rawsize;      /* Size of the data */
-        int             fd;           /* socket the request came from */
+	int             fd;           /* socket the request came from */
 	struct sockaddr_in srv_addr;  /* Server address */
 	struct sockaddr_in addr;      /* Remote party address */
 	REQUEST         *orig;        /* Original request. For proxy */
@@ -118,44 +118,44 @@ struct request {
 /* Request class structure
  */
 typedef struct request_class {
-        char *name;           /* Class name */
-        int  max_requests;    /* Max.number of pending requests of this type */
-        int  ttl;             /* Request time-to-live */
-        int  cleanup_delay;   /* Delay before cleaning the completed request */
+	char *name;           /* Class name */
+	int  max_requests;    /* Max.number of pending requests of this type */
+	int  ttl;             /* Request time-to-live */
+	int  cleanup_delay;   /* Delay before cleaning the completed request */
 	int  (*decode)(const struct sockaddr_in *srv_sa,
 		       const struct sockaddr_in *clt_sa,
 		       void *input, size_t inputsize, void **output);
-        int  (*respond)(REQUEST *r);          /* Handler function */
-        void (*xmit)(REQUEST *r);             /* Retransmit function */
-        int  (*comp)(void *a, void *b);       /* Compare function */
-        void (*free)(void *data);             /* Free the associated data */
-        void (*drop)(int type, void *data, void *old_data, int fd,
+	int  (*respond)(REQUEST *r);          /* Handler function */
+	void (*xmit)(REQUEST *r);             /* Retransmit function */
+	int  (*comp)(void *a, void *b);       /* Compare function */
+	void (*free)(void *data);             /* Free the associated data */
+	void (*drop)(int type, void *data, void *old_data, int fd,
 		     const char *msg);        /* Drop the request */
-        void (*cleanup)(int type, void *data);/* Cleanup function */
-        int (*failure)(int type, struct sockaddr_in *addr);
+	void (*cleanup)(int type, void *data);/* Cleanup function */
+	int (*failure)(int type, struct sockaddr_in *addr);
 	void (*update)(void *req, void *ptr);
 } REQUEST_CLASS;
 
 struct queue_stat {
-        size_t waiting;
-        size_t pending;
-        size_t completed;
+	size_t waiting;
+	size_t pending;
+	size_t completed;
 };
 typedef struct queue_stat QUEUE_STAT[R_MAX];
 
 typedef struct client {
 	grad_netdef_t           netdef;
-        char                    longname[GRAD_MAX_LONGNAME+1];
-        char                    *secret;
-        char                    shortname[GRAD_MAX_SHORTNAME+1];
+	char                    longname[GRAD_MAX_LONGNAME+1];
+	char                    *secret;
+	char                    shortname[GRAD_MAX_SHORTNAME+1];
 } CLIENT;
 
 typedef struct proxy_state {
 	uint32_t ref_ip;             /* Radius server IP */
-        uint32_t proxy_id;           /* Proxy ID assigned by the server */
-        uint32_t remote_ip;          /* Remote radius server IP */
-        uint32_t client_ip;          /* IP of the requesting client */ 
-        uint32_t id;                 /* Radius request ID */
+	uint32_t proxy_id;           /* Proxy ID assigned by the server */
+	uint32_t remote_ip;          /* Remote radius server IP */
+	uint32_t client_ip;          /* IP of the requesting client */
+	uint32_t id;                 /* Radius request ID */
 } PROXY_STATE;
 
 typedef struct {
@@ -169,13 +169,13 @@ typedef struct {
  * Internal representation of a user's profile
  */
 typedef struct user_symbol {
-        struct user_symbol *next;  /* Link to the next entry */
-        char *name;                /* Label */
-        grad_avp_t *check;         /* LHS */
-        grad_avp_t *reply;         /* RHS */
-        grad_locus_t loc;                 /* Location of the definition of
+	struct user_symbol *next;  /* Link to the next entry */
+	char *name;                /* Label */
+	grad_avp_t *check;         /* LHS */
+	grad_avp_t *reply;         /* RHS */
+	grad_locus_t loc;                 /* Location of the definition of
 				      this entry */
-        int ordnum;                /* Entry ordinal number (used for semantic
+	int ordnum;                /* Entry ordinal number (used for semantic
 				      checks */
 } User_symbol;
 
@@ -194,52 +194,52 @@ struct netname {
 
 typedef struct community Community;
 struct community {
-        char *name;
-        int access;
+	char *name;
+	int access;
 } ;
 
 typedef struct access_control_list ACL;
 struct access_control_list {
-        Community *community;   /* community or NULL to deny access */
+	Community *community;   /* community or NULL to deny access */
 	grad_list_t /* of grad_netdef_t */ *netlist;
 };
 
 struct radstat {
-        struct timeval start_time;
-        grad_counter_t port_active_count;
-        grad_counter_t port_idle_count;
+	struct timeval start_time;
+	grad_counter_t port_active_count;
+	grad_counter_t port_idle_count;
 };
 
 typedef enum {
-        port_idle = 1,
-        port_active
+	port_idle = 1,
+	port_active
 } port_status;
 
 typedef struct {
-        struct timeval start_time;
-        unsigned port_count; /* Number of ports in the port_stat array */
-        unsigned nas_count;  /* Number of NASes in the nas_stat array */
-        int nas_index; /* Next available NAS index */
-        Auth_server_stat auth;
-        Acct_server_stat acct;
-        /* struct nas_stat naslist[nas_count];
+	struct timeval start_time;
+	unsigned port_count; /* Number of ports in the port_stat array */
+	unsigned nas_count;  /* Number of NASes in the nas_stat array */
+	int nas_index; /* Next available NAS index */
+	Auth_server_stat auth;
+	Acct_server_stat acct;
+	/* struct nas_stat naslist[nas_count];
 	   struct port_stat portlist[port_count]; */
 } Server_stat;
 
 #define stat_inc(m,a,c) \
  do if (server_stat) {\
-        grad_nas_t *nas;\
-        server_stat -> m . c ++;\
-        if ((nas = grad_nas_lookup_ip(a)) != NULL && nas->app_data)\
-                ((struct nas_stat*)nas->app_data)-> m . c ++;\
+	grad_nas_t *nas;\
+	server_stat -> m . c ++;\
+	if ((nas = grad_nas_lookup_ip(a)) != NULL && nas->app_data)\
+		((struct nas_stat*)nas->app_data)-> m . c ++;\
  } while (0)
 
 extern struct radstat radstat;
 
 typedef struct snmp_req {
-        struct snmp_pdu *pdu;
-        char *community;
-        int access;
+	struct snmp_pdu *pdu;
+	char *community;
+	int access;
 	struct sockaddr_in addr;
 } SNMP_REQ;
 
@@ -266,7 +266,7 @@ enum auth_status {
 	auth_valid,           /* Authentication passed, expiration timeout
 				 is returned */
 	auth_account_expired, /* Account has expired */
-	auth_password_expired,/* Password has expired */ 
+	auth_password_expired,/* Password has expired */
 	auth_fail,
 	auth_nouser,
 	auth_reject,
@@ -345,8 +345,8 @@ extern char *server_id;
 extern Server_stat *server_stat;
 extern struct cfg_stmt snmp_stmt[];
 #endif
-extern int auth_comp_flag; 
-extern int acct_comp_flag; 
+extern int auth_comp_flag;
+extern int acct_comp_flag;
 extern RADIUS_USER exec_user;
 extern RADIUS_USER radiusd_user;
 extern grad_symtab_t *user_tab;
@@ -365,7 +365,7 @@ void input_register_method(INPUT *input,
 int input_register_channel(INPUT *input, char *name, int fd, void *data);
 void input_close_channels(INPUT *input);
 void input_close_channel_fd(INPUT *input, int fd);
-void input_close_channel_data(INPUT *input, char *name, void *data); 
+void input_close_channel_data(INPUT *input, char *name, void *data);
 int input_select(INPUT *input, struct timeval *tv);
 int input_select_channel(INPUT *input, char *name, struct timeval *tv);
 void *input_find_channel(INPUT *input, char *name, void *data);
@@ -441,7 +441,7 @@ extern struct cfg_stmt filters_stmt[];
 void guile_init(void);
 void scheme_load_path(char *pathname);
 int scheme_auth(char *procname, radiusd_request_t *req,
-                grad_avp_t *user_check, grad_avp_t **user_reply_ptr);
+		grad_avp_t *user_check, grad_avp_t **user_reply_ptr);
 int scheme_try_auth(int auth_type, radiusd_request_t *req,
 		    grad_avp_t *user_check,
 		    grad_avp_t **user_reply_ptr);
@@ -587,7 +587,7 @@ grad_nas_t *findnasbyindex(int);
 char *make_server_ident(void);
 void dump_users_db(void);
 void strip_username(int do_strip, char *name,
-                    grad_avp_t *check_item, char *stripped_name);
+		    grad_avp_t *check_item, char *stripped_name);
 int exec_program_wait (radiusd_request_t *request, grad_avp_t *rhs,
 		       grad_avp_t **reply, grad_avp_t **pfailed);
 
@@ -642,17 +642,18 @@ void snmp_req_free(void *ptr);
 void snmp_req_drop(int type, void *data, void *orig_data,
 		   int fd, const char *status_str);
 int snmp_req_respond(REQUEST *request);
-        
+
 /* radutil.c */
 radiusd_request_t *radiusd_request_alloc(grad_request_t *req);
 void radiusd_request_free(radiusd_request_t *radreq);
 
-char *radius_xlate(struct obstack *obp, char *str,
-                   grad_request_t *req, grad_avp_t *reply_pairs);
+void radius_strbuf_xlate(grad_strbuf_t obp, char const *str,
+			 grad_request_t *req, grad_avp_t *reply);
+char *radius_xlate(char const *str, grad_request_t *req, grad_avp_t *reply_pairs);
 int radius_eval_avp(radiusd_request_t *req, grad_avp_t *p, grad_avp_t *reply,
 		    int allow_xlat);
 int radius_eval_avl(radiusd_request_t *req, grad_avp_t *p);
-char *util_xlate(struct obstack *sp, char *fmt, grad_request_t *radreq);
+char *util_xlate(grad_strbuf_t sp, char const *fmt, grad_request_t *radreq);
 
 /* rewrite.y */
 extern struct cfg_stmt rewrite_stmt[];
@@ -703,18 +704,18 @@ void dynload_init(void);
 typedef struct channel Channel;
 
 struct channel {
-        char *name;
-        int  pmask[GRAD_LOG_NCAT]; 
-        int mode;   /* LM_ constant */
-        union {
+	char *name;
+	int  pmask[GRAD_LOG_NCAT];
+	int mode;   /* LM_ constant */
+	union {
 		struct {
 			int prio;   /* syslog priority */
 			int fac;    /* syslog facility */
 			char *tag;  /* syslog tag */
 		} sl;
-                char *file;      /* file: output file name */
-        } id;
-        int options;
+		char *file;      /* file: output file name */
+	} id;
+	int options;
 	char *prefix_hook;       /* prefix hook function */
 	char *suffix_hook;       /* suffix hook function */
 };
@@ -770,4 +771,3 @@ SCM sql_run_query(int type, const char *query);
 
 /* tsh.c */
 void tsh(void);
-
